"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const yargs_1 = __importDefault(require("yargs"));
const mongocluster_1 = require("./mongocluster");
const os_1 = __importDefault(require("os"));
const path_1 = __importDefault(require("path"));
const fs_1 = require("fs");
const mongodb_1 = require("mongodb");
const child_process_1 = require("child_process");
const debug_1 = __importDefault(require("debug"));
const events_1 = require("events");
(async function () {
    var _a;
    const defaultRunnerDir = path_1.default.join(os_1.default.homedir(), '.mongodb', 'runner2');
    const argv = await yargs_1.default
        .version(false)
        .scriptName('mongodb-runner')
        .option('topology', {
        alias: 't',
        choices: ['standalone', 'replset', 'sharded'],
        default: 'standalone',
        describe: 'Select a topology type',
    })
        .option('arbiters', {
        type: 'number',
        describe: 'number of arbiter nodes for each replica set',
    })
        .option('secondaries', {
        type: 'number',
        describe: 'number of secondaries for each replica set',
    })
        .option('shards', {
        type: 'number',
        describe: 'number of shards for sharded clusters',
    })
        .option('version', {
        type: 'string',
        describe: 'MongoDB server version to use',
    })
        .option('logDir', {
        type: 'string',
        describe: 'Directory to store server log files in',
    })
        .option('binDir', {
        type: 'string',
        describe: 'Directory containing mongod/mongos binaries',
    })
        .option('tmpDir', {
        type: 'string',
        default: os_1.default.tmpdir(),
        describe: 'Directory for temporary files',
    })
        .option('runnerDir', {
        type: 'string',
        default: defaultRunnerDir,
        describe: 'Directory for storing cluster metadata',
    })
        .option('docker', {
        type: 'string',
        describe: 'Docker image name to run server instances under',
    })
        .option('id', {
        type: 'string',
        describe: 'ID to save the cluster metadata under',
    })
        .option('all', {
        type: 'boolean',
        describe: 'for `stop`: stop all clusters',
    })
        .option('debug', { type: 'boolean', describe: 'Enable debug output' })
        .command('start', 'Start a MongoDB instance')
        .command('stop', 'Stop a MongoDB instance')
        .command('ls', 'List currently running MongoDB instances')
        .command('exec', 'Run a process with a MongoDB instance (as MONGODB_URI env var)')
        .demandCommand(1, 'A command needs to be provided')
        .help().argv;
    const [command, ...args] = argv._.map(String);
    if (argv.debug) {
        debug_1.default.enable('mongodb-runner');
    }
    async function start() {
        const id = argv.id || new mongodb_1.BSON.UUID().toHexString();
        if (!/^[a-zA-Z0-9_-]+$/.test(id)) {
            throw new Error(`ID '${id}' contains non-alphanumeric characters`);
        }
        await fs_1.promises.mkdir(argv.runnerDir, { recursive: true });
        const cluster = await mongocluster_1.MongoCluster.start({
            ...argv,
            args,
        });
        const serialized = await cluster.serialize();
        const { connectionString } = cluster;
        await fs_1.promises.writeFile(path_1.default.join(argv.runnerDir, `m-${id}.json`), JSON.stringify({ id, serialized, connectionString }));
        console.log(`Server started and running at ${connectionString}`);
        console.log('Run the following command to stop the instance:');
        console.log(`${argv.$0} stop --id=${id}` +
            (argv.runnerDir !== defaultRunnerDir
                ? `--runnerDir=${argv.runnerDir}`
                : ''));
        cluster.unref();
    }
    async function* instances() {
        for await (const { name } of await fs_1.promises.opendir(argv.runnerDir)) {
            if (name.startsWith('m-') && name.endsWith('.json')) {
                try {
                    const filepath = path_1.default.join(argv.runnerDir, name);
                    const stored = JSON.parse(await fs_1.promises.readFile(filepath, 'utf8'));
                    yield { ...stored, filepath };
                }
                catch (_a) {
                }
            }
        }
    }
    async function stop() {
        if (!argv.id && !argv.all) {
            throw new Error('Need --id or --all to know which server to stop');
        }
        const toStop = [];
        for await (const instance of instances()) {
            if (instance.id === argv.id || argv.all)
                toStop.push(instance);
        }
        await Promise.all(toStop.map(async ({ id, filepath, serialized, connectionString }) => {
            await (await mongocluster_1.MongoCluster.deserialize(serialized)).close();
            await fs_1.promises.rm(filepath);
            console.log(`Stopped cluster '${id}' (was running at '${connectionString}')`);
        }));
    }
    async function ls() {
        for await (const { id, connectionString } of instances()) {
            console.log(`${id}: ${connectionString}`);
        }
    }
    async function exec() {
        let mongodArgs;
        let execArgs;
        const doubleDashIndex = args.indexOf('--');
        if (doubleDashIndex !== -1) {
            mongodArgs = args.slice(0, doubleDashIndex);
            execArgs = args.slice(doubleDashIndex + 1);
        }
        else {
            mongodArgs = [];
            execArgs = args;
        }
        const cluster = await mongocluster_1.MongoCluster.start({
            ...argv,
            args: mongodArgs,
        });
        try {
            const [prog, ...progArgs] = execArgs;
            const child = (0, child_process_1.spawn)(prog, progArgs, {
                stdio: 'inherit',
                env: {
                    ...process.env,
                    MONGODB_URI: cluster.connectionString,
                    MONGODB_URL: cluster.connectionString,
                    MONGODB_HOSTPORT: cluster.hostport,
                },
            });
            [process.exitCode] = await (0, events_1.once)(child, 'exit');
        }
        finally {
            await cluster.close();
        }
    }
    async function unknown() {
        throw new Error(`Unknown command: ${command}. See '${argv.$0} --help' for more information.`);
    }
    await ((_a = { start, stop, exec, ls }[command]) !== null && _a !== void 0 ? _a : unknown)();
})().catch((err) => {
    process.nextTick(() => {
        throw err;
    });
});
//# sourceMappingURL=cli.js.map